/***************************************************************************
 * Copyright (c)  2014  Zephyr Software LLC. All rights reserved.
 *
 * This software is furnished under a license and/or other restrictive
 * terms and may be used and copied only in accordance with such terms
 * and the inclusion of the above copyright notice. This software or
 * any other copies thereof may not be provided or otherwise made
 * available to any other person without the express written consent
 * of an authorized representative of Zephyr Software LCC. Title to,
 * ownership of, and all rights in the software is retained by
 * Zephyr Software LCC.
 *
 * Zephyr Software LLC. Proprietary Information
 *
 * Unless otherwise specified, the information contained in this
 * directory, following this legend, and/or referenced herein is
 * Zephyr Software LLC. (Zephyr) Proprietary Information. 
 *
 * CONTACT
 *
 * For technical assistance, contact Zephyr Software LCC. at:
 *      
 *
 * Zephyr Software, LLC
 * 2040 Tremont Rd
 * Charlottesville, VA 22911
 *
 * E-mail: jwd@zephyr-software.com
 **************************************************************************/

#ifndef dollop_h
#define dollop_h

#include <assert.h>
#include <list>

namespace Zipr_SDK
{
class DollopManager_t;

class Placeable_t {
	public:
		Placeable_t() : m_is_placed(false), m_placed_address(0) {}
		void Place(Zipr_SDK::RangeAddress_t place) {
				assert(!m_is_placed);
				m_is_placed = true;
				m_placed_address = place;
		};
		Zipr_SDK::RangeAddress_t Place() { return m_placed_address; }
		bool IsPlaced() const { return m_is_placed; }
		friend std::ostream &operator<<(std::ostream &, const Placeable_t &);
	protected:
		bool m_is_placed;
		Zipr_SDK::RangeAddress_t m_placed_address;
};

class Dollop_t;

class DollopPatch_t : public Placeable_t {
	public:
		DollopPatch_t(Dollop_t *target) : m_target(target) {};
		DollopPatch_t() : m_target(NULL) { };

		Dollop_t *Target() const { return m_target; }
		void Target(Dollop_t *target) { m_target = target; }

		friend std::ostream &operator<<(std::ostream &, const DollopPatch_t &);
	private:
		Dollop_t *m_target;
};

class DollopEntry_t : public Placeable_t {
	public:
		DollopEntry_t(libIRDB::Instruction_t *insn, Dollop_t *member_of);
		libIRDB::Instruction_t *Instruction() const { return m_instruction; }
		void TargetDollop(Dollop_t *target) { m_target_dollop = target; }
		Dollop_t *TargetDollop() const { return m_target_dollop; }
		Dollop_t *MemberOfDollop() const { return m_member_of_dollop; }
		void MemberOfDollop(Dollop_t *member_of) { m_member_of_dollop = member_of; }

		bool operator==(const DollopEntry_t &);
		bool operator!=(const DollopEntry_t &);
		friend std::ostream &operator<<(std::ostream &, const DollopEntry_t &);
	private:
		libIRDB::Instruction_t *m_instruction;
		Dollop_t *m_target_dollop, *m_member_of_dollop;
};

class Dollop_t : public Placeable_t, public std::list<DollopEntry_t*> {
	public:
		static Dollop_t *CreateNewDollop(libIRDB::Instruction_t *start, 
		                                 DollopManager_t *mgr = NULL);

		Dollop_t(libIRDB::Instruction_t *start, DollopManager_t *);
		Dollop_t() :
			m_size(0),
			m_fallthrough_dollop(NULL),
			m_fallback_dollop(NULL),
			m_fallthrough_patched(false),
			m_coalesced(false),
			m_was_truncated(false),
			m_dollop_mgr(NULL) {}

		void SetDollopManager(DollopManager_t *mgr) { m_dollop_mgr = mgr; }

		size_t GetSize() const {
			return m_size;
		}
		void SetSize(size_t size) {
			m_size = size;
		}

		size_t GetDollopEntryCount() const {
			return size();
		}

		Dollop_t *Split(libIRDB::Instruction_t *split_point);
		void RemoveDollopEntries(std::list<DollopEntry_t*>::iterator,
		                         std::list<DollopEntry_t*>::iterator);

		friend std::ostream &operator<<(std::ostream &, const Dollop_t &);

		void FallbackDollop(Dollop_t *fallback) {
			m_fallback_dollop = fallback;
		}
		Dollop_t *FallbackDollop(void) const { return m_fallback_dollop; }

		void FallthroughDollop(Dollop_t *fallthrough) {
			m_fallthrough_dollop = fallthrough;
		}
		Dollop_t *FallthroughDollop(void) const { return m_fallthrough_dollop; }
		bool FallthroughPatched(void) const { return m_fallthrough_patched; }
		void FallthroughPatched(bool patched);

		DollopEntry_t *FallthroughDollopEntry(DollopEntry_t *) const;

		void WasTruncated(bool truncated) { m_was_truncated = truncated; }
		bool WasTruncated(void) const { return m_was_truncated; }

		void WasCoalesced(bool coalesced);
		bool WasCoalesced(void) const { return m_coalesced; }

		void ReCalculateSize();
	private:
		size_t CalculateSize();
		size_t m_size;
		Dollop_t *m_fallthrough_dollop, *m_fallback_dollop;
		bool m_fallthrough_patched;
		bool m_coalesced;
		bool m_was_truncated;
		DollopManager_t *m_dollop_mgr;
};
}
#endif
